//----------------------------------------------------------------
//
//  Fichero 		: MicroCPP
//	Tipo Componente	: Implementacin
//	Sistema		    : ANSI C++
//	Descripcin		: Procesador RISC en C++
//	Proyecto		: Programacin en C++
//
//----------------------------------------------------------------

#include <vector>
#include <iostream>
using namespace std;

//#define TRAZA

//************************************************************
//
//     		D E F I N I C I O N E S
//
//************************************************************

// Estado de las comparaciones
enum TEstado { noestado, menor, igual, mayor };



// Posicin de instrucciones (Contador de instrucciones)
typedef unsigned int TPosicion;

// Tipo de Valor de los operandos 
typedef int TValor;

// Formato de los operandos = 16 bits
struct TFormatoOp {
	unsigned int Codigo : 3;		// 3 bits para cdigo
	int  Valor  : 13;				// 13 bits para Valor
};

// Formato de las instrucciones
struct TFormatoInst {
 	unsigned int Codigo : 8;  // 8 bits para cdigo instrucciones
	TFormatoOp op1;
	TFormatoOp op2;
};

#define REGISTROS	10		// 10 registros
#define MEMORIA		1000	// 1000 posiciones de memoria

//************************************************************
//
//     		E X C E P C I O N E S
//
//************************************************************
class ExcepcionInterprete {
public:
	ExcepcionInterprete(const char* msg): 
		descripcion(msg) {
	}
	const char* Descripcion() {
		return descripcion;
	}
private:
	const char *descripcion;
};

class FueraRangoAlmacen : public ExcepcionInterprete {
public:
	FueraRangoAlmacen() : ExcepcionInterprete("Fuera de rango en almacen") {};
};

class FormatoInstruccionInvalido : public ExcepcionInterprete {
public:
	FormatoInstruccionInvalido() : ExcepcionInterprete("Formato de instruccion invalido") {};
};

class FormatoOperandoInvalido : public ExcepcionInterprete {
public:
	FormatoOperandoInvalido() : ExcepcionInterprete("Formato de operando invalido") {};
};

class OperacionInvalida : public ExcepcionInterprete {
public:
	OperacionInvalida(const char*s) : ExcepcionInterprete(s) {};
};


//************************************************************
//
//         D E F I N I C I O N    M E M O R I A    Y   E / S
//
//************************************************************

//************************************************************
//        Almacen DE DATOS GENERAL
//************************************************************

template <class T>
class Almacen {
public:
 	// Contruye el almacen de tamao n y con un valor inicial
 	Almacen(int n, T valorInicial) {
		Tam = n;
		pValores = new T[n];
		for (int i = 0; i < Tam; i++)
		pValores[i] = valorInicial;
	};

	~Almacen() {
		delete [] pValores;
	}

     // Lee de la posicin i
	T Leer( int i) throw(FueraRangoAlmacen)  {
		if ((i < 0) || (i >= Tam))
			throw FueraRangoAlmacen();
		return pValores[i];
	}

     // Escribe en una posicin un valor
	void Escribir(int i, T v) throw(FueraRangoAlmacen) {
		if ((i < 0) || ( i >= Tam))
			throw FueraRangoAlmacen();
		else
			pValores[i] = v;
	}

private:
	TValor *pValores;
	int Tam;	// Tamao del almacn
};

typedef Almacen<TValor> Registros;
typedef Almacen<TValor> Memoria;


//************************************************************
//   Pantalla DE SALIDA
//************************************************************

class Pantalla {
public:
	void Escribir(TValor v);
};

// Simplemente sacamos el valor por pantalla (cout)
void Pantalla::Escribir(TValor v) {
	cout << endl << v;
};


//************************************************************
//   ENTRADA
//************************************************************

class Teclado {
public:
	TValor Leer();
};

// Lee el valor desde el teclado (cin)
TValor Teclado::Leer() {
	TValor v;
	cin >> v;
    return v;
};




class ContenedorOperandos;
class ContenedorInstrucciones;
class Instruccion;
class Operando;
class Programa;

//************************************************************
//
//     		I N T E R P R E T E
//
//************************************************************

class Interprete {
public:
	Interprete (unsigned NumRegistros, unsigned TamMemoria)  {
        // Inicializa variables
		Estado = noestado;
		ContadorInstrucciones = 0;
		fin = false;
		pCI = NULL;
		pCO = NULL;
		pRegistros = new Registros(NumRegistros,0);	
		pMemoria = new Memoria(TamMemoria,0);	
		pTeclado = new Teclado;
		pPantalla = new Pantalla;

	}

	~Interprete() {
		delete pPantalla;
		delete pTeclado;
		delete pRegistros;
		delete pMemoria;
		if (pCI != NULL) delete pCI;
		if (pCO != NULL) delete pCO;
	}

	// Actualiza el estado para la comparacin
	void PonerEstado(TEstado e) {
		Estado = e;
	}

    // Lee el estado para el salto
	TEstado LeerEstado() {
		return Estado;
	}

	// Cambiar el puntero de instrucciones para el salto
	void CambiarPuntero(TPosicion p) {
		ContadorInstrucciones = p;
	}

	// Fin del programa -> Parar de interpretar
	void Fin() {
		fin = true;
	}

	void CargarPrograma(TFormatoInst codigo[]);

    // Bucle de interpretacin de instrucciones
	void Interpretar();


	ContenedorOperandos *ObtenerOperandos() {
		return pCO;
	}

	// Acceso a elemento globales

	Memoria *ObtenerMemoria() {
		return pMemoria;
	}
	Registros *ObtenerRegistros() {
		return pRegistros;
	}

	Pantalla *ObtenerPantalla() {
		return pPantalla;
	}

	Teclado *ObtenerTeclado() {
		return pTeclado;
	}

private:
	Programa *pPrograma;
	ContenedorInstrucciones *pCI;
	ContenedorOperandos *pCO;
	Pantalla *pPantalla;
	Teclado	*pTeclado;
	TPosicion ContadorInstrucciones;
	TEstado Estado;
	bool fin;	// Indicador de fin de programa
	// Implementamos los registros y memoria como almacn
	Registros    *pRegistros;	
	Memoria    *pMemoria;	

};






//************************************************************
//
//     		O P E R A N D O S
//
//************************************************************


//************************************************************
//     CLASE BASE DE LOS OPERANDOS
//************************************************************

class Operando {
public:
	Operando(Interprete *i) {
		pInterprete = i;
	}

	// Copia el formato del operando para poder acceder a l	
	PonFormato( TFormatoOp & f) {
		formato = f;
	};

	// Nos devuelve el valor del operando         
	virtual TValor Leer() = 0;

	// Escribe el valor del operando
	virtual void Escribir( TValor v) throw(OperacionInvalida) = 0 ;

    // Suma el operando op2 y escribe el resultado
	void operator += (Operando &op2);

    // Resta el operando op2 y escribe el resultado
	void operator -= (Operando &op2);

    // Compara op1 con op2
	friend TEstado Comparar (Operando &op1, Operando &op2);

	friend ostream& operator << (ostream&, const Operando&);

	virtual void Out(ostream&) = 0;

protected:
 	TFormatoOp formato;
	Interprete *pInterprete;
};

// Definimos el operando +=
void Operando::operator += (Operando &op2) {
	Escribir( Leer() + op2.Leer() );
};

// Definimos el operando -=
void Operando::operator -= (Operando &op2) {
	Escribir( Leer() - op2.Leer() );
};

ostream& operator << (ostream& s, const Operando&c) {
	c.Out(s);
	return s;
}

// Definimos Comparar
TEstado Comparar(Operando &op1, Operando &op2) {
	TValor v1 = op1.Leer(), v2 = op2.Leer();
	if (v1 > v2 )
		return mayor;
	else if (v1 < v2)
		return menor;
	else
		return igual;
};

//************************************************************
//     DEFINIMOS LAS CLASES DE LOS OPERANDOS
//************************************************************

// Definimos clase operando inmediato
class Inmediato:public Operando {
public:
	Inmediato(Interprete *i) : Operando(i) {
	}
	TValor Leer();
	void Escribir(TValor v) throw(OperacionInvalida);

	virtual void Out(ostream&);

};

TValor Inmediato::Leer() {
	return formato.Valor;
}

void Inmediato::Escribir(TValor ) throw(OperacionInvalida) {

	throw OperacionInvalida("No se puede escribir en operando inmediato");
};

void Inmediato::Out(ostream& sout) {
	sout << "#" << Leer();
}

// Definimos clase operando memoria
class OpMemoria:public Operando {
public:
	OpMemoria(Interprete *i) : Operando(i) {
	}
	TValor Leer();
	void Escribir(TValor v) throw(OperacionInvalida);
	virtual void Out(ostream&);
};

TValor OpMemoria::Leer() {
	return pInterprete->ObtenerMemoria()->Leer(formato.Valor);
}

void OpMemoria::Escribir(TValor v) throw(OperacionInvalida){
	pInterprete->ObtenerMemoria()->Escribir(formato.Valor,v);
};

void OpMemoria::Out(ostream& sout) {
	sout << "[" << formato.Valor << "]=" << pInterprete->ObtenerMemoria()->Leer(formato.Valor);
}


// Definimos clase operando registro
class OpRegistro:public Operando {
public:
	OpRegistro(Interprete *i) : Operando(i) {
	}
	TValor Leer();
	void Escribir(TValor v) throw(OperacionInvalida);
	virtual void Out(ostream&);
};

TValor OpRegistro::Leer() {

	return pInterprete->ObtenerRegistros()->Leer(formato.Valor);
}

void OpRegistro::Escribir(TValor v) throw(OperacionInvalida) {
	pInterprete->ObtenerRegistros()->Escribir(formato.Valor,v);
};

void OpRegistro::Out(ostream& sout) {
	sout << "<" << formato.Valor << ">=" << pInterprete->ObtenerRegistros()->Leer(formato.Valor);
}




class ContenedorOperandos {
private:
	vector<Operando*> VectorOperandos;

public:
	// Crea las intrucciones
	ContenedorOperandos (Interprete *i) : VectorOperandos(4) {
		VectorOperandos[0] = NULL;
		VectorOperandos[1] = new Inmediato(i);
		VectorOperandos[2] = new OpMemoria(i);
		VectorOperandos[3] = new OpRegistro(i);
	}

	//Elimina las instrucciones
	~ContenedorOperandos () {
		for (unsigned i = 1; i < VectorOperandos.size(); i++) {
			delete VectorOperandos[i];
		}
	}


	Operando * Obtener( TFormatoOp f) throw(FormatoOperandoInvalido) {
		Operando * rOperando;		
		try {
			rOperando = VectorOperandos.at(f.Codigo);
			rOperando->PonFormato(f);
		} catch (out_of_range e) {
			throw FormatoOperandoInvalido();
		}
		return rOperando;
	};

};



//************************************************************
//
//     		I N T R U C C I O N E S
//
//************************************************************


//************************************************************
//     CLASE BASE DE LAS INSTRUCCIONES
//************************************************************

class Instruccion {
public:                       	
	Instruccion(Interprete *i) {
		pInterprete = i;
	}

	// Guarda el formato de la intruccion
	void PonFormato (TFormatoInst f) {
		formato = f;
	}

    // Funcin virtual pura para ejecutar la instruccin
	virtual void Ejecutar() = 0;

protected:
	TFormatoInst formato;	// Formato de la instruccin
	Interprete *pInterprete;	// Guarda un puntero al pInterprete
};

//************************************************************
//     DEFINIMOS LAS CLASES DE LAS INTRUCCIONES
//************************************************************

class Mueve:public Instruccion {

public:
	Mueve(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Mueve::Ejecutar() {
	Operando *op1, *op2;

    // Coge los operandos
	op1 = pInterprete->ObtenerOperandos()->Obtener(formato.op1);
	op2 = pInterprete->ObtenerOperandos()->Obtener(formato.op2);

	// Escribe en el op1 el valor leido de op2
	op1->Escribir(op2->Leer());

	#ifdef TRAZA 
	cout << "op1 = " << *op2;
	#endif

}

class Suma:public Instruccion {

public:
	Suma(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Suma::Ejecutar() {

	Operando *op1,*op2;

	// Coge operandos
	op1 = pInterprete->ObtenerOperandos()->Obtener(formato.op1);
	op2 = pInterprete->ObtenerOperandos()->Obtener(formato.op2);

	#ifdef TRAZA 
	cout << "op1 = " << *op1 << " + " << *op2;
	#endif

    // Sumamos a op1, op2
	*op1 += *op2;

}

class Resta:public Instruccion {
public:
	Resta(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Resta::Ejecutar() {
	Operando *op1, *op2;

	// Coge operandos
	op1 = pInterprete->ObtenerOperandos()->Obtener( formato.op1);
	op2 = pInterprete->ObtenerOperandos()->Obtener( formato.op2);

	#ifdef TRAZA 
	cout << "op1 = " << *op1 << " - " << *op2;
	#endif


    // Restamos a op1, op2
	*op1 -= *op2;
}

class Compara:public Instruccion {
public:
	Compara(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Compara::Ejecutar() {
	Operando *op1, *op2;

    // Coge operandos
	op1 = pInterprete->ObtenerOperandos()->Obtener( formato.op1 );
	op2 = pInterprete->ObtenerOperandos()->Obtener( formato.op2 );

	#ifdef TRAZA 
	cout << "? " << *op1 << " >= " << *op2;
	#endif


    // Compara y cambia el estado del intrprete
	pInterprete->PonerEstado(Comparar(*op1, *op2));

}

class Salta:public Instruccion {
public:
	Salta(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Salta::Ejecutar() {
	Operando *op1;

	// Coge operando
	op1 = pInterprete->ObtenerOperandos()->Obtener(formato.op1);

    // Comprueba que la ltima comparacin es mayor o igual
	if ((pInterprete->LeerEstado() == mayor ) ||
		(pInterprete->LeerEstado() == igual )) {
            // Salta al valor del primer operando
		pInterprete->CambiarPuntero((TPosicion)op1->Leer());
		#ifdef TRAZA 
		cout << "SI-> " << *op1;
		#endif

	}
	else {
		#ifdef TRAZA 
		cout << "NO ";
		#endif
	}

}

class Entrada:public Instruccion {
public:
	Entrada(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Entrada::Ejecutar() {
	Operando *op1;

	// Coge operando
	op1 = pInterprete->ObtenerOperandos()->Obtener( formato.op1);

	#ifdef TRAZA 
	cout << ">> " << *op1 ;
	#endif

    // Escribe en el operando el valor ledo desde teclado
	op1 -> Escribir(pInterprete->ObtenerTeclado()->Leer());

}

class Salida:public Instruccion {
public:
	Salida(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Salida::Ejecutar() {
	Operando *op1;

	// Coge operando
	op1 = pInterprete->ObtenerOperandos()->Obtener(formato.op1);

	#ifdef TRAZA 
	cout << "<< " << *op1 ;
	#endif

	// Escribe en pantalla el valor del operando
	pInterprete->ObtenerPantalla()->Escribir(op1->Leer());

}

class Fin:public Instruccion {
public:
	Fin(Interprete *i) : Instruccion(i) {
	}
	virtual void Ejecutar();
};

void Fin::Ejecutar() {
	pInterprete->Fin();
}

class ContenedorInstrucciones {
private:
	vector<Instruccion*> VectorInstrucciones;

public:
	// Crea las intrucciones
	ContenedorInstrucciones(Interprete *i) : VectorInstrucciones(9) {
		VectorInstrucciones[0] = NULL;
		VectorInstrucciones[1] = new Mueve(i);
		VectorInstrucciones[2] = new Suma(i);
		VectorInstrucciones[3] = new Resta(i);
		VectorInstrucciones[4] = new Compara(i);
		VectorInstrucciones[5] = new Salta(i);
		VectorInstrucciones[6] = new Entrada(i);
		VectorInstrucciones[7] = new Salida(i);
		VectorInstrucciones[8] = new Fin(i);
	}

	//Elimina las instrucciones
	~ContenedorInstrucciones() {
		for (unsigned i = 1; i < VectorInstrucciones.size(); i++) {
			delete VectorInstrucciones[i];
		}
	}


	Instruccion * Obtener( TFormatoInst f) throw(FormatoInstruccionInvalido) {
		Instruccion * rInstruccion;		
		try {
			rInstruccion = VectorInstrucciones.at(f.Codigo);
			rInstruccion->PonFormato(f);
		} catch (out_of_range e) {
			throw FormatoInstruccionInvalido();
		}
		return rInstruccion;;
	};

};




//************************************************************
//
//     		P R O G R A M A
//
//************************************************************

// Programa ejemplo a ejecutar
// Saca por pantalla la suma acumulada de 1 a 10
TFormatoInst codigo[] = {
	{ 1, { 3, 0 }, { 1, 1 } },	//   0  MUEVE	R0,1
    { 1, { 2, 5 }, { 1, 0 } },	//   1  MUEVE	M5,0
	{ 2, { 2, 5 }, { 3, 0 } },	//   2  SUMA	M5,R0
	{ 7, { 2, 5 }, { 0, 0 } },	//   3  SALIDA	M5
	{ 2, { 3, 0 }, { 1, 1 } },	//   4  SUMA	R0,1      	  		
	{ 4, { 1,10 }, { 3, 0 } },	//   5  COMPARA	10,R0
	{ 5, { 1, 2 }, { 0, 0 } },	//   6  SALTA	@2
	{ 8, { 0, 0 }, { 0, 0 } }	//   7  FIN
};     	

class Programa {
public:
	// Carga el programa para su ejecucin
	Programa(TFormatoInst Programa[]) {
		pListadoPrograma = Programa;
	};

	TFormatoInst LeerInstruccion( TPosicion p) {
		// Lee el formato de la instruccion en la posicin p
	         return pListadoPrograma[p];
	};

private:
	TFormatoInst *pListadoPrograma;

};



//************************************************************
//
//     		I N T  R P R E T E
//
//************************************************************

void Interprete::CargarPrograma(TFormatoInst codigo[]) {
	cout << "Cargando programa..." << endl;
	pPrograma = new Programa(codigo);	

}




void Interprete::Interpretar() {
	Instruccion *I;

	cout << "Crea contenedores" << endl;

	// Crea los contenedores de instrucciones y operandos
	pCO = new ContenedorOperandos(this);
	pCI = new ContenedorInstrucciones(this);

	cout << "Empieza a interpretar..." << endl;
	while (!fin) {

		try {
			// Resuelve el formato y obtiene la instruccion
       		I = pCI->Obtener(pPrograma->LeerInstruccion(ContadorInstrucciones));
			#ifdef TRAZA 
			cout << endl << ContadorInstrucciones << ":" << typeid(*I).name() << "\t";
			#endif

			ContadorInstrucciones++;
			// Ejecuta la instruccin
			I->Ejecutar();
		}
		catch (ExcepcionInterprete e) {
			// Acaba el Interprete
			clog << "Error en la interpretacin: " << e.Descripcion() << " en linea " << ContadorInstrucciones << endl;
			fin = true;
		}
	}
}


//************************************************************
//
//     		M A I N
//
//************************************************************

void main () {
	cout << "Microprocesador RISC en C++" << endl;

	Interprete *it = new Interprete(10,1000);

	// Carga el programa
	it->CargarPrograma(codigo);

    // Lo interpreta
	it->Interpretar();

	delete it;
};
